// laser_mouse.c

/* ************************************************************** */
/*
    laser_mouse
    read from video4linux, filter pointing laser, tell X Mouseactions
    (ideas and code used from lasertraq-0.2, gqcam-0.9, sdlcam-0.8.0, tvtime-1.0.2)
    Copyright (C) 2007  Fred Koegler

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
Controls

Keys:
  ESCAPE: Quit
  F12: save video picture to file
  KP8 or UP: decrease video brightness with 5/256
  KP2 or DOWN: lower video brightness with 5/256
  F8: switch pointer interactionmode on/off
  KP_ENTER or RETURN: switch pointer interaction on/off
  KP_PLUS or PLUS: decrease polarisation value with 2
  KP_MINUS or MINUS: lower polarisation value with 2

  MousePressed: x,y coordinates printed to stderr
*/

#include "laser_mouse.h"

Display *x_display;
struct timeval button_down_time;

static int timediff( struct timeval *large, struct timeval *small )
{
  int tdiff = (large->tv_sec-small->tv_sec)*1000*1000
             + large->tv_usec-small->tv_usec;
  return(tdiff >= 0 ? tdiff : -tdiff );
}

void set_usecs_in(struct timeval *start_time)
{
  gettimeofday(start_time,0);
}

int get_usecs_since(struct timeval *start_time)
{
  struct timeval now;
  gettimeofday(&now,0);
  return timediff(&now,start_time);
}

#include "laser_mouse_video.c"

/* ************************************************************** */
void read_pref_file(FILE *input, struct VideoPaint *vp)
{
  char buff[1024], command[1024];
  int val=0;

  while(fgets(buff, sizeof(buff), input) != NULL)
  {
    if((buff[0] != '#') && (buff[0] != '\n'))
    {
      sscanf(buff," %s %d",command,&val);
      if(!strcmp(command,"video_device"))
        sscanf(buff," %s %s",command,vp->option.video_device);
      else if(!strcmp(command,"video_channel"))        vp->option.video_channel = val;
      else if(!strcmp(command,"video_brightness"))     vp->option.video_brightness = val;
      else if(!strcmp(command,"video_contrast"))       vp->option.video_contrast = val;
      else if(!strcmp(command,"video_color"))          vp->option.video_color = val;
      else if(!strcmp(command,"video_hue"))            vp->option.video_hue = val;
      else if(!strcmp(command,"video_whiteness"))      vp->option.video_whiteness = val;
      else if(!strcmp(command,"filter_autobright"))    vp->option.filter_autobright = val;
      else if(!strcmp(command,"filter_polarize"))      vp->option.filter_polarize = val;
      else if(!strcmp(command,"filter_polarize_show")) vp->option.filter_polarize_show = val;
      else if(!strcmp(command,"filter_swaprgb"))       vp->option.filter_swaprgb = val;
      else if(!strcmp(command,"savetype"))             vp->option.savetype = val;
      else if(!strcmp(command,"savetype_jpg_optimize"))
        vp->option.savetype_jpg_optimize = val;
      else if(!strcmp(command,"savetype_jpg_quality"))
        vp->option.savetype_jpg_quality = val;
      else if(!strcmp(command,"savetype_jpg_smoothness"))
        vp->option.savetype_jpg_smoothness = val;
      else if(!strcmp(command,"savetype_png_compression"))
        vp->option.savetype_png_compression = val;
      else if(!strcmp(command,"savetype_png_interlace"))
        vp->option.savetype_png_interlace = ( val==0 ? PNG_INTERLACE_NONE : PNG_INTERLACE_ADAM7 );
      else if(!strcmp(command,"savetype_ppm_format"))
        vp->option.savetype_ppm_format = val;
      else if(!strcmp(command,"vid_winxy0"))
        sscanf(buff," %s %u,%u",command,&vp->option.vid_winx0,&vp->option.vid_winy0);
      else if(!strcmp(command,"vid_winxy1"))
        sscanf(buff," %s %u,%u",command,&vp->option.vid_winx1,&vp->option.vid_winy1);
      else if(!strcmp(command,"capturexy0"))
        sscanf(buff," %s %u,%u",command,&vp->option.capturex0,&vp->option.capturey0);
      else if(!strcmp(command,"capturexy1"))
        sscanf(buff," %s %u,%u",command,&vp->option.capturex1,&vp->option.capturey1);
      else if(!strcmp(command,"mousexy0"))
        sscanf(buff," %s %u,%u",command,&vp->option.mousex0,&vp->option.mousey0);
      else if(!strcmp(command,"mousexy1"))
        sscanf(buff," %s %u",command,&vp->option.mousex1,&vp->option.mousey1);
      else if(!strcmp(command,"mousedown_starttime_max"))
        vp->option.mousedown_starttime_max = val;
      else fprintf(stderr,"Error: Unknown preference: %s\n", command);
    }
  }
}

/* ************************************************************** */
void save_pref_file(FILE *output, struct VideoPaint *vp)
{
  fprintf(output, "video_device %s\n",            vp->option.video_device);
  fprintf(output, "video_channel %d\n",           vp->option.video_channel);
  fprintf(output, "video_brightness %d\n",        vp->option.video_brightness);
  fprintf(output, "video_color %d\n",             vp->option.video_color);
  fprintf(output, "video_contrast %d\n",          vp->option.video_contrast);
  fprintf(output, "video_hue %d\n",               vp->option.video_hue);
  fprintf(output, "video_whiteness %d\n",         vp->option.video_whiteness);
  fprintf(output, "filter_autobright %d\n",       vp->option.filter_autobright);
  fprintf(output, "filter_polarize %d\n",         vp->option.filter_polarize);
  fprintf(output, "filter_polarize_show %d\n",    vp->option.filter_polarize_show);
  fprintf(output, "filter_swaprgb %d\n",          vp->option.filter_swaprgb);
  fprintf(output, "savetype %d\n",                vp->option.savetype);
  fprintf(output, "savetype_jpg_optimize %d\n",   vp->option.savetype_jpg_optimize);
  fprintf(output, "savetype_jpg_quality %d\n",    vp->option.savetype_jpg_quality);
  fprintf(output, "savetype_jpg_smoothness %d\n", vp->option.savetype_jpg_smoothness);
  fprintf(output, "savetype_png_compression %d\n",vp->option.savetype_png_compression);
  fprintf(output, "savetype_png_interlace %d\n",
    (vp->option.savetype_png_interlace==PNG_INTERLACE_NONE ? 0 : 1 ) );
  fprintf(output, "savetype_ppm_format %d\n",     vp->option.savetype_ppm_format);
  fprintf(output, "vid_winxy0 %u,%u\n",      vp->option.vid_winx0,vp->option.vid_winy0);
  fprintf(output, "vid_winxy1 %u,%u\n",      vp->option.vid_winx1,vp->option.vid_winy1);
  fprintf(output, "capturexy0 %u,%u\n",      vp->option.capturex0,vp->option.capturey0);
  fprintf(output, "capturexy1 %u,%u\n",      vp->option.capturex1,vp->option.capturey1);
  fprintf(output, "mousexy0 %u,%u\n",        vp->option.mousex0,vp->option.mousey0);
  fprintf(output, "mousexy1 %u,%u\n",        vp->option.mousex1,vp->option.mousey1);
  fprintf(output, "mousedown_starttime_max %u\n", vp->option.mousedown_starttime_max/1000);
}

static char short_options[] = "hVTf:F:r:v:x:d:b:c:o:p:z:u:w:a:s:t:O:Q:S:C:I:P:0:1:m:";
static struct option long_options[] =
{
  { "help", no_argument, NULL, 'h' },
  { "version", no_argument, NULL, 'V' },
  { "test", no_argument, NULL, 'T' },
  { "fullscreen", no_argument, NULL, 'f'},
  { "file", required_argument, NULL, 'F' },
  { "resolution", required_argument, NULL, 'r'},
  { "video_device", required_argument, NULL, 'v' },
  { "bright", required_argument, NULL, 'b' },
  { "contrast", required_argument, NULL, 'c' },
  { "color", required_argument, NULL, 'o' },
  { "hue", required_argument, NULL, 'u' },
  { "white", required_argument, NULL, 'w' },
  { "channel", required_argument, NULL, 'x' },
  { "filter_autobright", required_argument, NULL, 'a' },
  { "filter_polarize", required_argument, NULL, 'p' },
  { "filter_polarize_show", required_argument, NULL, 'z' },
  { "filter_swaprgb", required_argument, NULL, 's' },
  { "dump_file", required_argument, NULL, 'd' },
  { "type", required_argument, NULL, 't' },
  { "jpg_optimize", required_argument, NULL, 'O' },
  { "jpg_quality", required_argument, NULL, 'Q' },
  { "jpg_smoothness", required_argument, NULL, 'S' },
  { "png_compression", required_argument, NULL, 'C' },
  { "png_interlace", required_argument, NULL, 'I' },
  { "ppm_format", required_argument, NULL, 'P' },
  { "vid_winxy", required_argument, NULL, '0' },
  { "capturexy", required_argument, NULL, '1' },
  { "mousexy", required_argument, NULL, '2' },
  { "mousedown_starttime", required_argument, NULL, 'm' },
  { 0, 0, 0, 0 }
};

/* ************************************************************** */
void print_usage(void)
{
  fprintf(stderr,"%s - %s - %s\n",PROG,PROGRAM,VERSION);
  fprintf(stderr,"This software comes with ABSOLUTELY NO WARRANTY\n");
  fprintf(stderr,"This software is free software, and you are welcome to redistribute it\n");
  fprintf(stderr,"under certain conditions\n");
  fprintf(stderr,"usage: %s [-hVTF OPTIONS] [OPTIONS]\n",PROG);
  fprintf(stderr,"\t-h --help\t\tdisplay this help screen\n");
  fprintf(stderr,"\t-V --version\t\tdisplay version and exit\n");
  fprintf(stderr,"\t-T --test\tTest Video-Screen\n");
  fprintf(stderr,"\t-F --file <filename>\tread options from file <filename>\n\t\t\t\t(or stdin if filename is \"-\")\n");
  fprintf(stderr,"\t-f --fullscreen\tRun fullscreen\n");
  fprintf(stderr,"\t-r --resolution <wdith>x<height>\tResolution <wdith>x<height>\n");
  fprintf(stderr,"\t-v --video <video dev>\tgrab frames from <video dev>\n");
  fprintf(stderr,"\t-x --channel <num>\tChannel <num>\n");
  fprintf(stderr,"\t-b --bright <num>\tset brightness to [0,255]\n");
  fprintf(stderr,"\t-c --contrast <num>\tset contrast to [0,255]\n");
  fprintf(stderr,"\t-o --color <num>\tset color to [0,255]\n");
  fprintf(stderr,"\t-u --hue <num>\tset hue to [0,255]\n");
  fprintf(stderr,"\t-w --white <num>\tset white balance to [0,255]\n");
  fprintf(stderr,"\t-a --filter_autobright <num>\tperform autobrightness [0,1]\n");
  fprintf(stderr,"\t-p --filter_polarize <num>\tpolarization [0,255]\n");
  fprintf(stderr,"\t-z --filter_polarize_show <num>\tshow polarization in whole video [0,1]\n");
  fprintf(stderr,"\t-s --filter_swap <num>\t\tswap RGB into BGR [0,1]\n");
  fprintf(stderr,"\t-d --dump <filename>\tdump image to <filename>\n\t\t\t\t(or stdout if filename is \"-\")\n");
  fprintf(stderr,"\t-t --type <type>\ttype of image to dump(default:PNG)\n\t\t\t\t(valid types: PPM PNG JPEG)\n");
  fprintf(stderr,"\t-O --jpg_optimize <num>\t[0,1]\n");
  fprintf(stderr,"\t-Q --jpg_quality <num>\t[0,100]\n");
  fprintf(stderr,"\t-S --jpg_smoothness <num>\t[0,100]\n");
  fprintf(stderr,"\t-C --png_compression <num>\t[0,9]\n");
  fprintf(stderr,"\t-I --png_interlace <num>\t[0,1]\n");
  fprintf(stderr,"\t-P --ppm_format\t(valid types: RAW ASCII)\n");
  fprintf(stderr,"\t-0 --vid_winxy <x0>,<y0>,<x1>,<y1>\t\n");
  fprintf(stderr,"\t-1 --capturexy <x0>,<y0>,<x1>,<y1>\t\n");
  fprintf(stderr,"\t-2 --mousexy <x0>,<y0>,<x1>,<y1>\t\n");
  fprintf(stderr,"\t-m --mousedown_starttime <num>\tMousePressed <num> ms after detected light\n");
}

/* ************************************************************** */
int parameters(struct VideoPaint *vp, int *argc, char *argv[])
{
  int done = 0;
  int c;
  FILE *preffile;

  while (! done)
  {
    c = getopt_long(*argc,argv,short_options,long_options,NULL);
    switch (c) {
    case 'h' :
      print_usage();
      exit(0);
      break;
    case 'V' :
      fprintf(stdout,"%s - %s - %s\n",prog,program,version);
      exit(0);
      break;
    case 'T':
      vp->video_show = 1;
      break;
    case 'F' :
      vp->option.optionfile = 1;
      if(strcmp(optarg,"-") != 0) strcpy(vp->option.file,optarg);
      if (! strcmp(vp->option.file,"")) preffile = stdin;
      else if ((preffile = fopen(vp->option.file,"r")) == NULL)
      {
        fprintf(stderr,"Error: read %s\n",vp->option.file);
        exit(1);
      }
      read_pref_file(preffile,vp);
      fclose(preffile);
      break;
    case 'f':
      vp->screen.sdl_flags |= SDL_FULLSCREEN;
      break;
    case 'r':
      if (!optarg)
      {
        fprintf(stderr,"Error: resolution (<WIDTH>x<HEIGHT>)\n");
        exit(1);
      }
      else
      {
        sscanf(optarg,"%dx%d",&vp->screen.width,&vp->screen.height);
        if (vp->screen.width <= 0 || vp->screen.height <= 0)
        {
          fprintf(stderr,"Error: Invalid resolution\n");
          exit(1);
        }
      }
      break;
    case 'v' :
      sprintf(vp->option.video_device,"%s",optarg); break;
    case 'x' :
      vp->option.video_channel = atoi(optarg); break;
    case 'd' :
      vp->dump = 1; if(strcmp(optarg,"-") != 0) strcpy(vp->savefile,optarg); break;
    case 'b' :
      vp->option.video_brightness = atoi(optarg); break;
    case 'c' :
      vp->option.video_contrast = atoi(optarg); break;
    case 'o' :
      vp->option.video_color = atoi(optarg); break;
    case 'p' :
      vp->option.filter_polarize = atoi(optarg); break;
    case 'z' :
      vp->option.filter_polarize_show = atoi(optarg); break;
    case 'u' :
      vp->option.video_hue = atoi(optarg); break;
    case 'w' :
      vp->option.video_whiteness = atoi(optarg); break;
    case 'a':
      vp->option.filter_autobright = atoi(optarg); break;
    case 's':
      vp->option.filter_swaprgb = atoi(optarg); break;
    case 't' :
      if (!strcmp(optarg, "PPM"))        vp->option.savetype = PPM;
      else if (!strcmp(optarg, "PNG"))   vp->option.savetype = PNG;
      else if (!strcmp(optarg, "JPEG"))  vp->option.savetype = JPEG;
      break;
    case 'O':
      vp->option.savetype_jpg_optimize = atoi(optarg); break;
    case 'Q':
      vp->option.savetype_jpg_quality = atoi(optarg); break;
    case 'S':
      vp->option.savetype_jpg_smoothness = atoi(optarg); break;
    case 'C':
      vp->option.savetype_png_compression = atoi(optarg); break;
    case 'I':
      vp->option.savetype_png_interlace =
        (atoi(optarg) == 0 ? PNG_INTERLACE_NONE : PNG_INTERLACE_ADAM7);
      break;
    case 'P':
      if (!strcmp(optarg, "RAW"))        vp->option.savetype_ppm_format = RAW;
      else if (!strcmp(optarg, "ASCII")) vp->option.savetype_ppm_format = ASCII;
      break;
    case '0':
      if (4 != sscanf(optarg,"%u,%u,%u,%u",
        &vp->option.vid_winx0,&vp->option.vid_winy0,
        &vp->option.vid_winx1,&vp->option.vid_winy1))
      {
        fprintf(stderr,"Error: invalid option --vid_winxy \"x0,y0,x1,y1\"\n");
        exit(1);
      }
      break;
    case '1':
      if (4 != sscanf(optarg,"%u,%u,%u,%u",
        &vp->option.capturex0,&vp->option.capturey0,
        &vp->option.capturex1,&vp->option.capturey1))
      {
        fprintf(stderr,"Error: invalid option --capturexy \"x0,y0,x1,y1\"\n");
        exit(1);
      }
      break;
    case '2':
      if (4 != sscanf(optarg,"%u,%u,%u,%u",
        &vp->option.mousex0,&vp->option.mousey0,
        &vp->option.mousex1,&vp->option.mousey1))
      {
        fprintf(stderr,"Error: invalid option --mousexy \"x0,y0,x1,y1\"\n");
        exit(1);
      }
      break;
    case 'm':
      vp->option.mousedown_starttime_max = atoi(optarg); break;
    case '?' :
      fprintf(stderr,"Error: invalid option or ambiguous argument\n");
      print_usage();
      exit(1);
      break;
    case EOF :
      done = 1;
      break;
    default :
      print_usage();
      exit(1);
    };
  }
}

/* ************************************************************** */
int video_screen_init(struct VideoPaint *vp)
{
  int error;

  if ( SDL_Init(SDL_INIT_VIDEO) < 0 )
  {
    fprintf(stderr,"Error: sdl-init \"%s\"\n", SDL_GetError());
    error = -1;
  }
  else
  {
    atexit(SDL_Quit);

    if (vp->video_show)
    {
      if (! (vp->video.display = SDL_SetVideoMode(
        vp->video.width,vp->video.height,
        vp->video.depth,vp->video.sdl_flags)))
      {
        fprintf(stderr,"Error: sdl-display open \"%s\"\n",SDL_GetError());
        error = -2;
      }
      else
      {
        SDL_EventState(SDL_ACTIVEEVENT, SDL_IGNORE);
        vp->video.pixelbuf = (Uint8 *) vp->video.display->pixels;
        //SDL_ShowCursor(SDL_DISABLE);
        error = 0;
      }
    }
    else
    {
      if (! (vp->screen.display = SDL_SetVideoMode(
        vp->screen.width,vp->screen.height,
        vp->screen.depth,vp->screen.sdl_flags)))
      {
        fprintf(stderr,"Error: sdl-display open \"%s\"\n",SDL_GetError());
        error = -2;
      }
      else
      {
        vp->screen.pixelbuf = (Uint8 *) vp->screen.display->pixels;
        SDL_EventState(SDL_ACTIVEEVENT, SDL_IGNORE);
        error = 0;
      }
    }
  }
  return error;
}

/* ************************************************************** */
int screen_draw_frame(struct VideoPaint *vp)
{
  if (vp->video_show)
  {
    if (SDL_MUSTLOCK (vp->video.display))
      if (SDL_LockSurface (vp->video.display) < 0)
      {
        fprintf(stderr,"Error: Can't lock Surface: %s\n",SDL_GetError());
        return -1;
      }

    memcpy((char*)vp->video.pixelbuf,vp->pic,vp->video.size * 3);

    if (SDL_MUSTLOCK (vp->video.display))
      SDL_UnlockSurface(vp->video.display);

    SDL_UpdateRect(vp->video.display,0,0,vp->video.width,vp->video.height);
  }
  else
  {
    if (SDL_MUSTLOCK (vp->screen.display))
      if (SDL_LockSurface (vp->screen.display) < 0)
      {
        fprintf(stderr,"Error: Can't lock Surface: %s\n",SDL_GetError());
        return -1;
      }

    memset((char*)vp->screen.pixelbuf,(char)int(255*rand()),vp->screen.size * 3);

    if (SDL_MUSTLOCK (vp->screen.display))
      SDL_UnlockSurface(vp->screen.display);

    SDL_UpdateRect(vp->screen.display,0,0,vp->screen.width,vp->screen.height);
  }
  return 0;
}

int mouse_click(Display *x_display, int mousedown, int press,
  int move=0, int mouse_x=0, int mouse_y=0)
{
  XEvent event;

  memset(&event, 0, sizeof(event));
  long eventmask = 0L;

  if (move)
  {
    event.type = MotionNotify;
    event.xmotion.same_screen = True;
    event.xmotion.is_hint = NotifyNormal;
    if (mousedown)
    {
      event.xmotion.state = Button1MotionMask;
      eventmask |= Button1MotionMask;
    }
    else
    {
      event.xmotion.state = PointerMotionMask;
      eventmask |= PointerMotionMask;
    }
    fprintf(stderr,"click(): mouse-move\n");

    XWarpPointer(x_display,None,DefaultRootWindow(x_display),0,0,0,0,mouse_x,mouse_y);

//    XQueryPointer(x_display,
//      RootWindow(x_display,DefaultScreen(x_display)),
//      &event.xmotion.root,&event.xmotion.window,
//      &event.xmotion.x_root,&event.xmotion.y_root,
//      &event.xmotion.x,&event.xmotion.y,
//      &event.xmotion.state);

//    event.xmotion.subwindow = event.xmotion.window;

//    while (event.xmotion.subwindow)
//    {
//      event.xmotion.window = event.xmotion.subwindow;
//      XQueryPointer(x_display,event.xmotion.window,
//        &event.xmotion.root,&event.xmotion.subwindow,
//        &event.xmotion.x_root,&event.xmotion.y_root,
//        &event.xmotion.x,&event.xmotion.y,
//        &event.xmotion.state);
//    }
//    if (XSendEvent(x_display,PointerWindow,True,eventmask,&event) == 0)
//      fprintf(stderr, "Error: XSendEvent\n");
  }

  if (press != mousedown)
  {
    event.xbutton.button = Button1;
    event.xbutton.same_screen = True;
    event.xbutton.state = Button1Mask;
    if (press)
    {
      event.type = ButtonPress;
      eventmask |= ButtonPressMask;
      if (mousedown)
        fprintf(stderr,"click(): mouse still down\n");
      else
        fprintf(stderr,"click(): mouse-down\n");
      mousedown = 1;
    }
    else
    {
      if (mousedown)
      {
        event.type = ButtonRelease;
        eventmask |= ButtonReleaseMask;
        fprintf(stderr,"click(): mouse-up\n");
      }
    }
    XQueryPointer(x_display,
      RootWindow(x_display,DefaultScreen(x_display)),
      &event.xbutton.root,&event.xbutton.window,
      &event.xbutton.x_root,&event.xbutton.y_root,
      &event.xbutton.x,&event.xbutton.y,
      &event.xbutton.state);

    event.xbutton.subwindow = event.xbutton.window;

    while (event.xbutton.subwindow)
    {
      event.xbutton.window = event.xbutton.subwindow;
      XQueryPointer(x_display,event.xbutton.window,
        &event.xbutton.root,&event.xbutton.subwindow,
        &event.xbutton.x_root,&event.xbutton.y_root,
        &event.xbutton.x,&event.xbutton.y,
        &event.xbutton.state);
    }
  }

  if (XSendEvent(x_display,PointerWindow,True,eventmask,&event) == 0)
    fprintf(stderr, "Error: XSendEvent\n");

  XFlush(x_display);

  return (mousedown);
}

void mouse_move(Display *x_display, int mouse_x, int mouse_y, int mousedown)
{
//  XWarpPointer(x_display,None,DefaultRootWindow(x_display),
//    0,0,0,0,mouse_x,mouse_y);
//  if (mousedown) 
mouse_click(x_display,mousedown,mousedown,1,mouse_x,mouse_y);
//  XFlush(x_display);
}

/* ************************************************************** */
void MOUSEMOTION(struct VideoPaint *vp, int x, int y)
{
  vp->mouse_x = x;
  vp->mouse_y = y;
  if (! vp->mouse_button)
  {
    vp->old_x = vp->mouse_x;
    vp->old_y = vp->mouse_y;
  }
  mouse_move(x_display,vp->mouse_x,vp->mouse_y,vp->mouse_button);
}

/* ************************************************************** */
void MOUSEBUTTONUP(struct VideoPaint *vp)
{
  vp->mouse_button = mouse_click(x_display,vp->mouse_button,0);
}

/* ************************************************************** */
void MOUSEBUTTONDOWN(struct VideoPaint *vp)
{
  vp->mouse_button = mouse_click(x_display,0,1);
}

/* ************************************************************** */
int main(int argc, char *argv[])
{
  static struct VideoPaint vp;
  int i;
  unsigned char buff[3];
  char preffile_[1024];
  FILE *preffile;
  SDL_Event event;
  int mx,my,xy0,xy1;

  if((x_display = XOpenDisplay(NULL)) == NULL)
  {
    fprintf(stderr,"Error: Cannot open X-Display...\n");
    exit(-1);
  }

  vp.video_show = 0;

  vp.video.display = NULL;
  vp.video.width = 0;
  vp.video.height = 0;
  vp.video.depth = 24;
  vp.video.sdl_flags = 0;

  vp.active = 0;
  vp.mousedown = 0;
  vp.mousedown_start = 0;
  vp.option.mousedown_starttime_max = 1000;
  set_usecs_in(&vp.mousedown_starttime);
  vp.dev = 0;
  vp.dump = 0;
  vp.savefile[0] = 0;
  vp.pic = NULL;
  vp.pic_temp = NULL;

  strcpy(vp.option.video_device, "/dev/video");

  vp.option.optionfile = 0;
  vp.option.file[0] = 0;

  vp.option.filter_autobright = 0;
  vp.option.filter_swaprgb = 0;
  vp.option.filter_polarize = 0;
  vp.option.filter_polarize_show = 0;

  vp.option.video_channel = 0;
  vp.option.video_brightness = 128;
  vp.option.video_contrast = 128;
  vp.option.video_color = 128;
  vp.option.video_hue = 128;
  vp.option.video_whiteness = 128;

  vp.option.savetype = PNG;
  vp.option.savetype_jpg_smoothness = 0;
  vp.option.savetype_jpg_quality = 75;
  vp.option.savetype_jpg_optimize = 0;
  vp.option.savetype_png_compression = 6;
  vp.option.savetype_png_interlace = PNG_INTERLACE_NONE;
  vp.option.savetype_ppm_format = RAW;

  vp.option.vid_winxy = 0;
  vp.option.vid_winx0 = vp.option.vid_winy0 = 0;
  vp.option.vid_winx1 = vp.option.vid_winy1 = 0;
  vp.option.capturexy = 0;
  vp.option.capturex0 = vp.option.capturey0 = 0;
  vp.option.capturex1 = vp.option.capturey1 = 0;
  vp.option.mousexy = 0;
  vp.option.mousex0 = vp.option.mousey0 = 0;
  vp.option.mousex1 = 1024;
  vp.option.mousey1 = 768;

  vp.screen.width = 400;
  vp.screen.height = 64;
  vp.screen.sdl_flags = 0;

  vp.mouse_button = 0;

  if (strcmp("",getenv("HOME")))
    sprintf(preffile_,"%s/.%s.rc",getenv("HOME"),PROG);
  else
  {
    sprintf(preffile_,"%s.rc",PROG);
    fprintf(stderr,"Error: environment variable HOME = \"\" using \"%s\"\n",preffile_);
  }

  if ((preffile = fopen(preffile_,"r")) == NULL)
    fprintf(stderr,"Error: read %s\n",preffile_);
  else
  {
    read_pref_file(preffile,&vp);
    fclose(preffile);
  }

  parameters(&vp,&argc,argv);

  vp.option.vid_winxy = 0;
  if (vp.option.vid_winx0 | vp.option.vid_winy0
    | vp.option.vid_winx1 | vp.option.vid_winy1)
  {
    vp.option.vid_winxy = 1;
    xy0 = min(vp.option.vid_winx0,vp.option.vid_winx1);
    xy1 = max(vp.option.vid_winx0,vp.option.vid_winx1);
    vp.option.vid_winx0 = xy0;
    vp.option.vid_winx1 = xy1;
    xy0 = min(vp.option.vid_winy0,vp.option.vid_winy1);
    xy1 = max(vp.option.vid_winy0,vp.option.vid_winy1);
    vp.option.vid_winy0 = xy0;
    vp.option.vid_winy1 = xy1;
  }

  vp.option.capturexy = 0;
  if (vp.option.capturex0 | vp.option.capturey0
    | vp.option.capturex1 | vp.option.capturey1)
  {
    vp.option.capturexy = 1;
    xy0 = min(vp.option.capturex0,vp.option.capturex1);
    xy1 = max(vp.option.capturex0,vp.option.capturex1);
    vp.option.capturex0 = xy0;
    vp.option.capturex1 = xy1;
    xy0 = min(vp.option.capturey0,vp.option.capturey1);
    xy1 = max(vp.option.capturey0,vp.option.capturey1);
    vp.option.capturey0 = xy0;
    vp.option.capturey1 = xy1;
  }

  video_open_cam(&vp);
  video_get_cam_info(&vp,1);

  if (vp.vid_caps.channels > 1
    && vp.option.video_channel >= 0
    && vp.option.video_channel < vp.vid_caps.channels)
  {
    vp.vid_chan.channel = vp.option.video_channel;
    if (video_set_cam_vid_chan(&vp) < 0)
    {
      fprintf(stderr,"Error: Channel %d\n",vp.vid_chan.channel);
      exit(-1);
    }
  }

  vp.vid_pic.brightness = vp.option.video_brightness * 256;
  vp.vid_pic.colour = vp.option.video_color * 256;
  vp.vid_pic.contrast = vp.option.video_contrast * 256;
  vp.vid_pic.hue = vp.option.video_hue * 256;
  vp.vid_pic.whiteness = vp.option.video_whiteness * 256;

  if (vp.vid_pic.depth == 24) vp.vid_pic.palette = VIDEO_PALETTE_RGB24;
  else if (vp.vid_pic.depth == 16) vp.vid_pic.palette = VIDEO_PALETTE_UYVY;
//  else if (vp.vid_pic.depth == 16) vp.vid_pic.palette = VIDEO_PALETTE_YUV422;
  if (video_set_cam_vid_pic(&vp) < 0)
  {
    vp.vid_pic.depth = 24;
    vp.vid_pic.palette = VIDEO_PALETTE_RGB24;
    if (video_set_cam_vid_pic(&vp) < 0)
    {
      vp.vid_pic.depth = 16;
      vp.vid_pic.palette = VIDEO_PALETTE_UYVY;
      if (video_set_cam_vid_pic(&vp) < 0)
      {
        vp.vid_pic.depth = 16;
        vp.vid_pic.palette = VIDEO_PALETTE_YUV422;
        if (video_set_cam_vid_pic(&vp) < 0)
        {
          fprintf(stderr,"Error: RGB24 / UYVY / YUV422\n");
          exit(-1);
        }
      }
    }
  }

  if (vp.option.vid_winxy)
  {
    vp.vid_win.x = vp.option.vid_winx0;
    vp.vid_win.y = vp.option.vid_winy0;
    vp.vid_win.width = vp.option.vid_winx1-vp.option.vid_winx0;
    vp.vid_win.height = vp.option.vid_winy1-vp.option.vid_winy0;
  }
  else
  {
    vp.vid_win.x = 0;
    vp.vid_win.y = 0;
    vp.vid_win.width = vp.vid_caps.maxwidth;
    vp.vid_win.height = vp.vid_caps.maxheight;
  }

  vp.video.width = vp.vid_win.width;
  vp.video.height = vp.vid_win.height;

  if (video_set_cam_vid_win(&vp) < 0)
  {
    fprintf(stderr,"Error: Videocapture %dx%d\n",
      vp.vid_win.width,vp.vid_win.height);
    exit(1);
  }
  video_get_cam_info(&vp,0);
  video_print_cam_info(&vp);

  vp.video.rect.x = 0;
  vp.video.rect.y = 0;
  vp.video.rect.w = vp.video.width;
  vp.video.rect.h = vp.video.height;
  vp.video.size = vp.video.width * vp.video.height;
  vp.video.depth = 24;

  vp.screen.rect.x = 0;
  vp.screen.rect.y = 0;
  vp.screen.rect.w = vp.screen.width;
  vp.screen.rect.h = vp.screen.height;
  vp.screen.size = vp.screen.width*vp.screen.height;
  vp.screen.depth = 24;

  if ((vp.option.mousex0 | vp.option.mousey0
     | vp.option.mousex1 | vp.option.mousey1) == 0 )
  {
    vp.option.mousex1 = vp.screen.width;
    vp.option.mousey1 = vp.screen.height;
  }

  if (video_screen_init(&vp))
  {
    fprintf(stderr,"Error: sdl-init.\n");
    return (-1);
  }

  char *title = (char*)malloc (strlen(PROGRAM)+strlen(PROG)+strlen(VERSION)+10); 
  sprintf (title,"%s - %s - %s",PROG,PROGRAM,VERSION);
  SDL_WM_SetCaption(title, NULL);
  free (title);

  int rette_filter_polarize = vp.option.filter_polarize;
  long len;
  int SDL_Loop = 1;
  int old_palette = -1;
  vp.option.mousedown_starttime_max *= 1000;
  while (SDL_Loop)
  {
    SDL_Delay(20);

    if (vp.dev)
    {
      if (vp.vid_pic.palette == VIDEO_PALETTE_UYVY) /* 9-UYVY HAUPAUGE bt878 */
        len = read (vp.dev,vp.pic,vp.video.size*2);
      else if (vp.vid_pic.palette == VIDEO_PALETTE_YUV422) /* 7-YUV422 HAUPAUGE bt878 */
        len = read (vp.dev,vp.pic,vp.video.size*2);
      else len = read (vp.dev,vp.pic,vp.video.size*3);

      if (len <= 0) fprintf(stderr,"Error reading image...\n");
      else
      {
        if (vp.vid_pic.palette == VIDEO_PALETTE_UYVY) video_filter_uyvy2rgb(&vp);
        if (vp.vid_pic.palette == VIDEO_PALETTE_YUV422) video_filter_yuv4222rgb(&vp);
        if (vp.option.filter_swaprgb) video_filter_swaprgb(&vp);
        if (vp.option.filter_autobright) video_filter_autobright(&vp);
        if (vp.option.capturexy) video_filter_capturexy(&vp);
        if (vp.option.filter_polarize) video_filter_polarize(&vp);
        if (vp.dump)
        {
          SDL_Loop = 0;
          video_dump_pict(&vp);
        }
        else
          if (screen_draw_frame(&vp)) SDL_Loop = 0;
      }
    }

    if (SDL_Loop)
    {
      while (SDL_PollEvent(&event))
      {
        switch (event.type)
        {
        case SDL_QUIT:
          SDL_Loop = 0;
          break;

        case SDL_KEYUP:
          break;

        case SDL_KEYDOWN:
          switch (event.key.keysym.sym)
          {
          case SDLK_F12:
            if (! vp.savefile[0]) sprintf(vp.savefile,"./%s_pict",PROG);
            fprintf(stderr,"dump ... %s\n",vp.savefile);
            video_dump_pict(&vp);
            break;
          case SDLK_F8:
            if (vp.option.filter_polarize)
            {
              rette_filter_polarize = vp.option.filter_polarize;
              vp.option.filter_polarize = 0;
            }
            else
            {
              if (! rette_filter_polarize) rette_filter_polarize = 128;
              vp.option.filter_polarize = rette_filter_polarize;
            }
            break;
          case SDLK_KP_PLUS:
          case SDLK_PLUS:
            if (vp.option.filter_polarize)
              if (vp.option.filter_polarize < 254)
              {
                vp.option.filter_polarize += 2;
                rette_filter_polarize = vp.option.filter_polarize;
                fprintf(stderr,"Polarize: %d\n",vp.option.filter_polarize);
              }
            break;
          case SDLK_KP_MINUS:
          case SDLK_MINUS:
            if (vp.option.filter_polarize)
              if (vp.option.filter_polarize > 2)
              {
                vp.option.filter_polarize -= 2;
                rette_filter_polarize = vp.option.filter_polarize;
                fprintf(stderr,"Polarize: %d\n",vp.option.filter_polarize);
              }
            break;
          case SDLK_KP8:
          case SDLK_UP:
            if (vp.option.video_brightness)
              if (vp.option.video_brightness < 250)
              {
                vp.option.video_brightness += 5;
                fprintf(stderr,"Brightness: %d\n",vp.option.video_brightness);
                vp.vid_pic.brightness = vp.option.video_brightness * 256;
                if (video_set_cam_vid_pic(&vp) < 0)
                {
                  fprintf(stderr,"Error: ioctl (VIDIOCSPICT)\n");
                }
              }
            break;
          case SDLK_KP2:
          case SDLK_DOWN:
            if (vp.option.video_brightness)
              if (vp.option.video_brightness > 5)
              {
                vp.option.video_brightness -= 5;
                fprintf(stderr,"Brightness: %d\n",vp.option.video_brightness);
                vp.vid_pic.brightness = vp.option.video_brightness * 256;
                if (video_set_cam_vid_pic(&vp) < 0)
                {
                  fprintf(stderr,"Error: ioctl (VIDIOCSPICT)\n");
                }
              }
            break;
          case SDLK_ESCAPE:
            SDL_Loop = 0;
            break;
          case SDLK_KP_ENTER:
          case SDLK_RETURN:
            vp.active = ! vp.active;
            fprintf(stderr,"RETURN video.active %d\n",vp.active);
            break;
          default:
            break;
          }
          break; 

        case SDL_MOUSEMOTION:
          mx = (int)event.motion.x;
          my = (int)event.motion.y;
          break;

        case SDL_MOUSEBUTTONUP:
          break;

        case SDL_MOUSEBUTTONDOWN:
          fprintf(stderr,"Point: %d %d\n",mx,my);
          break;

        default:
          fprintf(stderr,"Unknown SDL event received ?!\n");
          break;
        }
      }
    }
  }

  if (! vp.option.optionfile)
  {
    if ((preffile = fopen(preffile_,"w")) == NULL)
      fprintf(stderr,"Fehler: save %s\n",preffile_);
    else
    {
      save_pref_file(preffile, &vp);
      fclose(preffile);
    }
  }

  video_close_cam(&vp,1);

  SDL_ShowCursor(SDL_ENABLE);

  exit(0);
}
